"""
    DSSC-detector class module
    --------------------------

    The dssc detector class. It represents a namespace for frequent evaluation
    while implicitly applying/requiring certain structure/naming conventions to
    its objects.

    comments:
        - contributions should comply with pep8 code structure guidelines.
        - Plot routines don't fit into objects since they are rather fluent.
          They have been outsourced to dssc_plot.py. They can now be accessed
          as tbdet member functions.
"""
import os
import logging
import joblib

import numpy as np
import xarray as xr

from ..load import load_run
from ..util.exceptions import ToolBoxValueError, ToolBoxFileError
from .dssc_data import (
        save_xarray, load_xarray, save_attributes_h5,
        search_files, get_data_formatted)
from .dssc_misc import (
        load_dssc_info, get_xgm_formatted)
from .dssc_geodat import (
    tile_geodat, module_geodat, quad_geodat)
from .dssc_processing import (
        process_dssc_data, create_empty_dataset)
from .dssc_geometry import convert_hex2cart

__all__ = ["DSSCBinner", "DSSCFormatter", "DSSC_Generator"]


log = logging.getLogger(__name__)


class DSSCBinner:
    def __init__(self, proposal_nr, run_nr,
                 binners={},
                 xgm_name='SCS_SA3', 
                 ):
        """
        A dssc binner object. Loads and bins the dssc data according to the
        bins specified in 'binners'. The data can be reduced further through
        masking using XGM or TIM data.

        Parameters
        ----------
        proposal_nr: int, str
            proposal number containing run folders
        run_nr: int, str
            run number
        binners: dictionary
            dictionary containing binners constructed using the
            'create_dssc_bins' tbdet-method.
        xgm_name: str
            a valid mnemonic key of the XGM data to be used to mask the dssc
            frames. Since the xgm is used in several methods its name can be
            set here globally.

        Returns
        -------
        DSSCbinner: object

        Example
        -------
        1.) quick -> generic bins, no xgm,
        >>> import toolbox_scs.detectors as tbdet
        >>> run235 = tbdet.DSSCBinner(proposal_nb=2212, run_nb=235)
        2.) detailed -> docs
        """

        # ---------------------------------------------------------------------
        # object (run) properties
        # ---------------------------------------------------------------------
        self.proposal = proposal_nr
        self.runnr = run_nr
        self.info = load_dssc_info(proposal_nr, run_nr)
        self.run = load_run(proposal_nr, run_nr)
        self.binners = {}
        for b in binners:
            self.add_binner(b, binners[b])
        self.xgm_name = xgm_name
        # ---------------------------------------------------------------------
        # Prepare pulse mask for additional data reduction next to binning
        # ---------------------------------------------------------------------
        self.xgm = None
        self.pulsemask = None

        log.debug("Constructed DSSC object")

    def __del__(self):
        pass

    def add_binner(self, name, binner):
        """
        Add additional binner to internal dictionary

        Parameters
        ----------
        name: str
            name of binner to be created
        binner: xarray.DataArray
            An array that represents a map how the respective coordinate should
            be binned.

        Raises
        ------
        ToolBoxValueError: Exception
            Raises exception in case the name does not correspond to a valid
            binner name. To be generalized.
        """
        if name in ['trainId', 'pulse', 'x', 'y']:
            self.binners[name] = binner
        else:
            msg = "Invalid binner name"
            log.info(msg+", no binner created")
            raise ToolBoxValueError(msg, name)

    def load_xgm(self, xgm_coord_stride=1):
        """
        load xgm data and construct coordinate array according to corresponding
        dssc frame number.
        """
        self.xgm = get_xgm_formatted(self.run, self.xgm_name, xgm_coord_stride)

    def create_xgm_mask(self,
                        xgm_threshold=(0, np.inf), normevery=1):
        """
        creates a mask for dssc frames according to measured xgm intensity.
        Once such a mask has been constructed, it will be used in the data
        reduction process to drop out-of-bounds pulses.
        """
        fpt = self.info['frames_per_train']
        n_trains = self.info['number_of_trains']
        trainIds = self.info['trainIds']
        data = np.ones([n_trains, fpt], dtype=bool)
        self.pulsemask = xr.DataArray(data,
                                      dims=['trainId', 'pulse'],
                                      coords={'trainId': trainIds,
                                              'pulse': range(fpt)})

        self.load_xgm(xgm_coord_stride=normevery)
        valid = (self.xgm > xgm_threshold[0]) * \
                (self.xgm < xgm_threshold[1])
        self.pulsemask = \
                (valid.combine_first(self.pulsemask).astype(bool))[:,0:fpt]

        log.info(f'created pulse mask used during processing')

    def get_info(self):
        """
        Returns the expected shape of the binned dataset, in case binners have
        been defined.
        """
        if any(self.binners):
            empty = create_empty_dataset(self.info, self.binners)
            return(empty.dims)
        else:
            log.info("no binner defined yet.")
            pass

    def get_xgm_binned(self):
        """
        Bin the xgm data according to the binners of the dssc data. The result
        can eventually be merged into the final dataset by the DSSCFormatter.

        Returns
        -------
        xgm_data: xarray.DataSet
            xarray dataset containing the binned xgm data
        """
        if self.xgm is not None:
            xgm_data = self.xgm.to_dataset(name='xgm')
            for b in self.binners:
                if b in ['trainId', 'pulse']:
                    xgm_data[b+"_binned"] = self.binners[b]
                    xgm_data = xgm_data.groupby(b+"_binned").mean(b)
                    xgm_data = xgm_data.rename(name_dict={b+"_binned": b})
            return xgm_data.transpose('trainId', 'pulse')
        else:
            log.warning("no xgm data. Use load_xgm() to load the xgm data.")
            pass

    # -------------------------------------------------------------------------
    # Data processing
    # -------------------------------------------------------------------------
    def process_data(self, modules=[], filepath='./',
                     chunksize=512, backend='loky', n_jobs=None,
                     dark_image=None,
                     xgm_normalization=False, normevery=1
                    ):
        """
        Load and bin dssc data according to self.bins. No data is returned by
        this method. The condensed data is written to file by the worker
        processes directly.

        Parameters
        ----------
        modules: list of ints
            a list containing the module numbers that should be processed. If
            empty, all modules are processed.
        filepath: str
            the path where the files containing the reduced data should be
            stored.
        chunksize: int
            The number of trains that should be read in one iterative step.
        backend: str
            joblib multiprocessing backend to be used. At the moment it can be
            any of joblibs standard backends: 'loky' (default),
            'multiprocessing', 'threading'. Anything else than the default is
            experimental and not appropriately implemented in the dbdet member
            function 'bin_data'.
        n_jobs: int
            inversely proportional of the number of cpu's available for one 
            job. Tasks within one job can grab a maximum of n_CPU_tot/n_jobs of 
            cpu's.
            Note that when using the default backend there is no need to adjust 
            this parameter with the current implementation.
        dark_image: xarray.DataArray
            DataArray with dimensions compatible with the loaded dssc data. If
            given, it will be subtracted from the dssc data before the binning.
            The dark image needs to be of dimension module, trainId, pulse, x 
            and y.
        xgm_normalization: boolean
            if true, the dssc data is normalized by the xgm data before the
            binning.
        normevery: int
            integer indicating which out of normevery frame will be normalized.
        """
        log.info("Bin data according to binners")
        log.info(f'Process {chunksize} trains per chunk')

        mod_list = modules
        if len(mod_list) == 0:
            mod_list = [i for i in range(16)]
        log.info(f'Process modules {mod_list}')
        njobs = n_jobs
        if njobs is None:
            njobs = len(mod_list)

        module_jobs = []
        for m in mod_list:
            dark = dark_image
            if dark_image is not None:
                dark = dark_image.sel(module=m)
            module_jobs.append(dict(
                proposal=self.proposal,
                run_nr=self.runnr,
                module=m,
                chunksize=chunksize,
                path=filepath,
                info=self.info,
                dssc_binners=self.binners,
                pulsemask=self.pulsemask,
                dark_image=dark,
                xgm_normalization=xgm_normalization,
                xgm_mnemonic=self.xgm_name,
                normevery=normevery,
            ))

        data = None

        log.info(f'using parallelization backend {backend}')
        joblib.Parallel(n_jobs=njobs, backend=backend) \
            (joblib.delayed(process_dssc_data)(**module_jobs[i]) \
             for i in range(len(mod_list)))

        log.info(f'Binning done')


class DSSCFormatter:
    def __init__(self, filepath):
        """
        Class that handles formatting related aspects, before handing the data
        overt for analysis.

        Parameters
        ----------
        filepath: str
            location of processed files.

        Raises
        ------
        ToolBoxFileError: Exception
            Trows an error in case the the given path does not exist.
        """
        self._filenames = []
        self._filename = ''
        if os.path.exists(filepath):
            try:
                self._filenames = search_files(filepath)
            except ToolBoxFileError:
                log.info("path did not contain any files")
        else:
            log.info("path did not exist")

        self.data = None
        self.data_xarray = {}
        self.attributes = {}

    def combine_files(self, filenames=[]):
        """
        Read the files given in filenames, and store the data in the class
        variable 'data'. If no filenames are given, it tries to read the files
        stored in the class-internal variable '_filenames'.

        Parameters
        ----------
        filenames: list
            list of strings containing the names of the files to be combined.
        """
        if any(filenames) is True:
            self._filenames = filenames
        if self._filenames is not None:
            self.data = get_data_formatted(self._filenames)
        else:
            log.info("No matching data found.")

    def add_dataArray(self, groups=[]):
        """
        Reads addional xarray-data from the first file given in the list of
        filenames. This assumes that all the files in the folder contain the
        same additional data. To be generalized.

        Parameters
        ----------
        groups: list
            list of strings with the names of the groups in the h5 file,
            containing additional xarray data.
        """
        if any(self._filenames) is True:
            for group in groups:
                self.data_xarray[group] = load_xarray(self._filenames[0],
                                                      group=group,
                                                      form='array')
        else:
            log.info("No files found in specified folder.")

    def add_attributes(self, attributes={}):
        """
        Add additional information, such as run-type, as attributes to the
        formatted .h5 file.

        Parameters
        ----------
        attributes: dictionary
            a dictionary, containing information or data of any kind, that
            will be added to the formatted .h5 file as attributes.
        """
        for key in attributes.keys():
            self.attributes[key] = attributes[key]

    def save_formatted_data(self, filename):
        """
        Create a .h5 file containing the main dataset in the group called
        'data'. Additional groups will be created for the content of the
        variable 'data_array'. Metadata about the file is added in the form of
        attributes.

        Parameters
        ----------
        filename: str
            the name of the file to be created
        """
        save_xarray(filename, self.data, group='data', mode='w')

        for arr in self.data_xarray.keys():
            save_xarray(filename, self.data_xarray[arr], group=arr, mode='a')

        save_attributes_h5(filename, self.attributes)


class DSSC_Generator:
    def __init__(self, q_shift = quad_geodat):
        # tile_geodat: [[voffset, hoffset], [is_flip, axis]]
        # axis= 0, flip up-down
        # axis= 1, flip left-right
        # fractional shifts not included so far
        self.tile_geodat = tile_geodat
        self.module_geodat = module_geodat
        self.quad_geodat = q_shift

    def read_geometry(self, filename):
        pass

    def save_geometry(self, filename):
        pass

    def full_detector(self, dssc_data):
        log.debug('start alignment of module data')
        # assemble full detector
        module_data = joblib.Parallel(n_jobs=16) \
            (joblib.delayed(self.convert_module)
                (n, dssc_data[n, :, :]) for n in range(16))

        log.debug('assemble converted module data in quadrants')
        q_data = []
        for i in range(4):
            i0 = i*4
            q_data.append(self.quadrant(i+1, module_data[i0:i0+4]))

        log.debug('align quadrants')
        # determine dimensions
        [qdimy, qdimx] = q_data[0].shape
        fdimy = np.max(
                    [np.abs(self.quad_geodat[1, 0]) + self.quad_geodat[1, 1],
                     np.abs(self.quad_geodat[3, 0]) + self.quad_geodat[3, 1]]
                    ) + 2 * qdimy
        fdimx = np.max(
                    [np.abs(self.quad_geodat[2, 0]) + self.quad_geodat[2, 1],
                     np.abs(self.quad_geodat[0, 0]) + self.quad_geodat[0, 1]]
                    ) + 2 * qdimx

        full_dssc = np.zeros((fdimy, fdimx))

        # quadrant 2
        y0 = self.quad_geodat[1, 0]
        y1 = qdimy+self.quad_geodat[1, 0]
        x0 = 0
        x1 = qdimx
        full_dssc[y0:y1, x0:x1] = q_data[1]

        # quadrant 1
        y0 = qdimy+self.quad_geodat[1, 0]+self.quad_geodat[1, 1]
        y1 = 2*qdimy+self.quad_geodat[1, 0]+self.quad_geodat[1, 1]
        x0 = self.quad_geodat[0, 0]
        x1 = self.quad_geodat[0, 0]+qdimx
        full_dssc[y0:y1, x0:x1] = q_data[0]

        # quadrant 4
        y0 = qdimy \
            + self.quad_geodat[1, 0] \
            + self.quad_geodat[1, 1] \
            + self.quad_geodat[3, 0]
        y1 = 2 * qdimy \
            + self.quad_geodat[1, 0] \
            + self.quad_geodat[1, 1] \
            + self.quad_geodat[3, 0]
        x0 = self.quad_geodat[0, 0] \
            + self.quad_geodat[0, 1] \
            + qdimx
        x1 = self.quad_geodat[0, 0] \
            + self.quad_geodat[0, 1] \
            + 2*qdimx
        full_dssc[y0:y1, x0:x1] = q_data[3]

        # quadrant 3
        y0 = self.quad_geodat[1, 0] \
            + self.quad_geodat[1, 1] \
            + self.quad_geodat[3, 0] \
            - self.quad_geodat[3, 1]
        y1 = qdimy + self.quad_geodat[1, 0] \
            + self.quad_geodat[1, 1] \
            + self.quad_geodat[3, 0] \
            - self.quad_geodat[3, 1]
        x0 = self.quad_geodat[0, 0] \
            + self.quad_geodat[0, 1] \
            + qdimx \
            + self.quad_geodat[2, 0]
        x1 = self.quad_geodat[0, 0] \
            + self.quad_geodat[0, 1] \
            + 2*qdimx \
            + self.quad_geodat[2, 0]
        full_dssc[y0:y1, x0:x1] = q_data[2]

        return full_dssc

    def quadrant(self, q, module_data):
        """
        q: quadrant index 0-3
        dssc_data: (15, 128, 512)
        """
        log.debug("assembling quadrant: Q" + str(q))
        if q == 1:
            module0 = module_data[0]
            mdimx = module0.shape[1]
            qu = [module0]
            for n, m in zip(range(1, 4), range(1, 4)):
                qu.append(np.zeros(
                            (self.module_geodat[n, 0], mdimx),
                            dtype='float32'))
                qu.append(module_data[m])

        elif q == 2:
            module0 = module_data[0]
            mdimx = module0.shape[1]
            qu = [module0]
            for n, m in zip(range(5, 8), range(1, 4)):
                qu.append(np.zeros(
                            (self.module_geodat[n, 0], mdimx),
                            dtype='float32'))
                qu.append(module_data[m])

        elif q == 3:
            module0 = module_data[3]
            mdimx = module0.shape[1]
            qu = [module0]
            for n, m in zip(range(10, 7, -1), range(2, -1, -1)):
                qu.append(np.zeros(
                            (self.module_geodat[n, 0], mdimx),
                            dtype='float32'))
                qu.append(module_data[m])

        elif q == 4:
            module0 = module_data[3]
            mdimx = module0.shape[1]
            qu = [module0]
            for n, m in zip(range(14, 11, -1), range(2, -1, -1)):
                qu.append(np.zeros(
                            (self.module_geodat[n, 0], mdimx),
                            dtype='float32'))
                qu.append(module_data[m])

        return np.vstack(qu)

    def convert_module(self, m, module_hex):
        """
        convert hex to cartesian
        add gap according to geometry file
        flip according to geometry file
        returns the module in cartesian coordinates with the correct gap
        between tiles
        """
        log.debug("processing module: " + str(m))
        tile_hex = self.split_module(module_hex)
        offset, flip_module = self._get_tile_geodat(m)

        module_cart = np.zeros(
                            (120 + offset[0], 2*276 + offset[1]),
                            dtype='float32')
        module_cart[0:120, 0:276] = convert_hex2cart(tile_hex[0, :, :])
        module_cart[0:120, 276 + offset[1]:552 + offset[1]] \
            = convert_hex2cart(tile_hex[1, :, :])

        if flip_module[0]:
            module_cart = np.flip(module_cart, axis=flip_module[1])

        return module_cart

    def split_module(self, module_hex):
        # return two tiles
        tile_hex = np.zeros((2, 128, 256), dtype='float32')
        tile_hex[0, :, :] = module_hex[:, 0:256]
        tile_hex[1, :, :] = module_hex[:, 256:512]
        return tile_hex

    def _get_tile_geodat(self, m):
        return self.tile_geodat[m, 0, :], self.tile_geodat[m, 1, :]

    def _default_geometry(self):
        pass

# for all modules do
# 1) fix pedestals by ASIC mask and add/subtract pedestal offset
# 2) split module into its tiles
# 3) convert tiles from hex to cart
# 4) reposition tiles according to offsets (from geometry file)
# 5) figure out orientation and flip up-down or left-right (flag in geometry
#    file)
