"""
    DSSC-geometry related sub-routines.

    comment: contributions should comply with pep8 code structure guidelines.
"""
import logging

import numpy as np
import math as math
import skimage.util as sku
import skimage.transform as skt

from ..misc.image_manipulation import pad4fft, get_im_center

log = logging.getLogger(__name__)


def n_choose_k(n, k):
    """
    :n choose k problem: returns the number of ways to choose a subset k from a
    set of n elements this is eqivalent to:
        factorial(n)/factorial(k)/factorial(n-k)

    written by A. Scherz
    """
    return math.factorial(n) / math.factorial(k) / math.factorial(n - k)


def thiran_type_I(order=3):
    return Thiran(1, order)


def thiran_type_II(order=2):
    return Thiran(2, order)


class Thiran:
    """
    there are two types of 1D fractional filters based on Thiran filters. both
    types can have an order N=1 to inf. some types, provide simple algorithms
    for applying the delay to a signal. for type II and orders larger than N.
    he transfer and impulse response fct have to be calculated and are more
    computational expensive
    The generator defaults the Thiran filter of type I and order N=3
    For more details see of the filters and algorithms,
        see pg. 685, IEEE Trans. on Im. Proc. 17, 679 (2008)

    written by A. Scherz, XFEL, 2012-11-26
    updated by A. Scherz, XFEL, 2019-02-11 for python
    added to SpeckleToolbox by A. Scherz, XFEL, 2020-08-06
    """

    type_I = staticmethod(thiran_type_I)

    type_II = staticmethod(thiran_type_II)

    def __init__(self, filter_type=1, order=3, delay=0):
        self._type = filter_type
        self._n = order
        self._delay = delay
        self._b = np.zeros(order)
        self._a_minus_delay = 0
        self._a_plus_delay = 0

    def __str__(self):
        print('Thiran 1D Fractional filter')
        print('  Type : ' + self._type.__str__())
        print(' Order : ' + self._n.__str__())
        print(' Delay : ' + self._delay.__str__())

        return ''

    def apply_delay(self, signal):
        # periodic boundaries are important. add padding of at least the order
        # N left and right
        # this should be applied before calling this method for type I
        shifted_signal = np.array(signal)

        if self._type == 1:     # type I Thiran
            for j in range(len(shifted_signal) - self._n - 1, self._n - 1, -1):
                for k in range(1, self._n + 1):
                    shifted_signal[j] += self._b[k - 1] \
                        * (shifted_signal[j - k] - shifted_signal[j + k])
        else:   # type II of order N = 2
            for j in range(self._n - 1, len(shifted_signal) - self._n - 1, 1):
                shifted_signal[j] += self._a_plus_delay \
                    * (shifted_signal[j - 1] - shifted_signal[j + 1])
            for j in range(len(shifted_signal) - self._n - 1, self._n - 1, -1):
                shifted_signal[j] += self._a_minus_delay \
                    * (shifted_signal[j + 1] - shifted_signal[j - 1])

        return shifted_signal  # shifted signal

    def delay(self, tau):
        if (tau < -0.5) or (tau > 0.5):
            raise ValueError('WARNING out of range: delay: 0.0 - 0.5')
        self._delay = tau
        # update coefficients for delay
        self._coefficients()
        return self

    def _coefficients(self):
        for k in range(1, self._n + 1):
            p_k = 1
            for l in range(self._n + 1):
                p_k *= (self._delay - l) / (self._delay - l - k)
            # n_choose_k function is factorial(n)/factorial(k)/factorial(n-k)
            self._b[k - 1] = (-1) ** k * p_k * n_choose_k(self._n, k)
        self._a_plus_delay = \
            (-4 + self._delay**2 + np.sqrt(12 - 3*self._delay**2))
        self._a_minus_delay = self._a_plus_delay
        self._a_plus_delay /= (3*self._delay + 2 + self._delay**2)
        self._a_minus_delay /= (-3*self._delay + 2 + self._delay**2)


def shear_im(im: np.array, delay: float,
             im_axis: int = 1,
             thiran: Thiran = thiran_type_I()) -> np.array:
    """
    : direction (str): 1=='hor' or 0=='ver', required
    : thiran: Thiran type 1 of order 3 is default (default)
    """
    im = np.float32(im)
    cx, cy = get_im_center(im)
    # print(im.shape)
    # print(cx, cy)
    if im_axis == 0:
        for k in range(2 * cy):
            shift = delay * (k - cy)
            # full pixel shifts
            im[:, k] = np.roll(im[:, k], round(shift))
            # now shift a pixel fraction
            im[:, k] = thiran.delay(shift-round(shift)).apply_delay(im[:, k])
            #  thiran_filter(im[:, k], thiran(order, shift - round(shift)))
        # print('apply vertical shear')
    elif im_axis == 1:
        for k in range(2 * cx):
            shift = delay * (k - cx)
            # full pixel shifts
            im[k, :] = np.roll(im[k, :], round(shift))
            # now shift a pixel fraction: shift between +/- [0-0.5]
            im[k, :] = thiran.delay(shift - round(shift)).apply_delay(im[k, :])
            # thiran_filter(im[k, :], thiran(order, shift - round(shift)))
        #print('apply horizontal shear')
    else:
        # TODO: throw exception
        print('direction undefined')

    return im


def convert_cart2hex(im, thiran=thiran_type_II()):
    """
    [ imH ] = ConversionCart2Hex( imC, orderN, showImage )
    imC: image in cartesian coord.
    thiran: Thiran filter type 1 of order 3 is default
    show: optional, if true an image representation in hex is shown.
    note: show hex image may take considerable amount of time for larger
    images!)
    Nov.28, 2012, A. Scherz
        """
    # first step: zero pad the image for extending boundaries

    im_hex = pad4fft(im, dim=[2*im.shape[0], 2*im.shape[1]], fill='reflect')
    cx, cy = get_im_center(im_hex)

    sqrt3 = math.sqrt(3)
    shear1 = sqrt3 - math.sqrt(6 / sqrt3)
    shear2 = math.sqrt(sqrt3 / 6)
    shear3 = 2 - math.sqrt(6 / sqrt3)
    # this is conversion from cartesian to hexagonal lattice
    im_hex = shear_im(im_hex, -shear1, im_axis=0, thiran=thiran)
    im_hex = shear_im(im_hex, -shear2, im_axis=1, thiran=thiran)
    im_hex = shear_im(im_hex, -shear3, im_axis=0, thiran=thiran)

    # reorder of pixels required
    height = int(np.round(cart2hex([im.shape[0], 0]))[0])
    width = int(np.round(cart2hex([0, im.shape[1]]))[1])
    y0 = height % 2

    # print(height, width, y0)
    for y in range(-height // 2, height // 2 + y0 + 1):
        im_hex[y - y0 + cx, :] = np.roll(im_hex[y - y0 + cx, :],
                                         math.floor((y - y0) / 2))

    # now crop image
    im_hex = im_hex[-height // 2 + cx: height // 2 + cx + y0,
                    -width // 2 + cy: width // 2 + cy + 1]

    # TODO check height and width
    # implement show

    return im_hex


def convert_hex2cart(im_hex, thiran=thiran_type_II()):
    """
    [ imH ] = ConversionCart2Hex( imC, orderN, showImage )
    imC: image in cartesian coord.
    thiran: Thiran filter type 1 of order 3 is default
    show: optional, if true an image representation in hex is shown.
    note: show hex image may take considerable amount of time for larger
    images!)
    Nov.28, 2012, A. Scherz
    """
    # first step: zero pad the image for extending boundaries

    im = pad4fft(im_hex,
                 dim=[2*im_hex.shape[0], 2*im_hex.shape[1]],
                 fill='reflect')
    cx, cy = get_im_center(im)

    # need to skew image to hexagonal shape before apply shear operations
    height = im_hex.shape[0]
    y0 = height % 2
    for y in range(- height // 2, height // 2 + y0 + 1):
        im[y - y0 + cx, :] = np.roll(im[y - y0 + cx, :],
                                     - math.floor((y + y0) / 2))

    sqrt3 = math.sqrt(3)
    shear1 = sqrt3 - math.sqrt(6 / sqrt3)
    shear2 = math.sqrt(sqrt3 / 6)
    shear3 = 2 - math.sqrt(6 / sqrt3)
    # this is conversion from cartesian to hexagonal lattice
    im = shear_im(im, shear3, im_axis=0, thiran=thiran)
    im = shear_im(im, shear2, im_axis=1, thiran=thiran)
    im = shear_im(im, shear1, im_axis=0, thiran=thiran)

    # the image should now form a rectangular box of following size:
    height = int(np.round(hex2cart([im_hex.shape[0], 0]))[0])
    width = int(np.round(hex2cart([0, im_hex.shape[1]]))[1])
    y0 = height % 2

#    print(height, width, y0)
#    for y in range(-height // 2, height // 2 + y0 + 1):
#        im_hex[y - y0 + cx, :] = np.roll(im_hex[y - y0 + cx, :],
#                                         math.floor((y - y0) / 2))

    # now crop image
    im = im[-height // 2 + cx: height // 2 + cx + y0,
            -width // 2 + cy: width // 2 + cy + 1]

    # TODO check height and width
    # implement show

    return im


def cart2hex(r_cart):
    # TODO: this function can eventually move somewhere else
    sqrt3 = math.sqrt(3)
    # conversion matrix
    rhex = math.sqrt(2 / sqrt3) * np.array([[1, 1 / 2], [0, sqrt3 / 2]])

    return np.dot(rhex, r_cart)


def hex2cart(r_hex):
    sqrt3 = math.sqrt(3)
    rhex = math.sqrt(2 / sqrt3) * np.array([[1, 1 / 2], [0, sqrt3 / 2]])
    rcart = np.linalg.inv(rhex)

    return np.dot(rcart, r_hex)


def slice_module(module):
    tile = np.zeros((2, 128, 256), dtype='float32')
    tile[0, :, :] = module[:, 0:256]
    tile[1, :, :] = module[:, 256:512]
    print(tile.shape)
    return tile


def data_from_symmetry(data, center=[584.0, 544.0], threshold=0.0):
    """
    Parameters
    ----------
    data: np.ndarray
        2-dimensional numpy-array containing the raw image data.
    center: list
        a list with two values indicating the physical center, i.e. the
        position of the beam.
    threshold: float
        The lower magnitude limit, below which data from the image should
        be replaced with data from the rotated image.

    Returns
    -------
    data_rotated: np.ndarray
        A 2-dimensional numpy-array containing data from the rotated image,
        where it is missing in the original one.
    """
    image = np.nan_to_num(sku.img_as_float(data))
    shape = np.shape(image)

    shifts = [(shape[i] - 2*(shape[i]/2.0 - center[i])) for i in range(2)]

    tform = skt.EuclideanTransform(rotation=np.pi, translation=shifts[::-1])
    data_rotated = skt.warp(image, tform)

    mask = (image <= threshold)
    data_rotated[~mask] = 0.0
    return data_rotated
